from os.path import join
from ament_index_python.packages import get_package_share_directory
from launch import LaunchDescription
from launch_ros.actions import Node
from launch.actions import ExecuteProcess
import yaml

def generate_launch_description():
    launch_root = get_package_share_directory('ar3_config')
    description_root = get_package_share_directory('ar3_description')
    config_root = join(launch_root, 'config')

    urdf_path = join(description_root, 'ar3.urdf')
    with open(urdf_path, 'r') as f:
        robot_description = {'robot_description': f.read()}

    srdf_path = join(description_root, 'ar3.srdf')
    with open(srdf_path, 'r') as f:
        robot_description_semantic = {'robot_description_semantic': f.read()}

    robot_description_kinematics = {}
    with open(join(config_root, 'kinematics.yaml'), 'r') as f:
        robot_description_kinematics.update({"robot_description_kinematics": yaml.safe_load(f)})

    # Planning Functionality
    ompl_planning_pipeline_config = {
        "move_group": {
            "planning_plugin": "ompl_interface/OMPLPlanner",
            "request_adapters": """default_planner_request_adapters/AddTimeOptimalParameterization default_planner_request_adapters/FixWorkspaceBounds default_planner_request_adapters/FixStartStateBounds default_planner_request_adapters/FixStartStateCollision default_planner_request_adapters/FixStartStatePathConstraints"""    ,
            "start_state_max_bounds_error": 0.1,
        }
    }

    with open(join(config_root, "ompl_planning.yaml"), 'r') as f:
       ompl_planning_pipeline_config["move_group"].update(yaml.safe_load(f))

     # Trajectory Execution Functionality
    moveit_controllers = {
        "moveit_controller_manager": "moveit_simple_controller_manager/MoveItSimpleControllerManager"
    }
    with open(join(config_root, "ar3_controllers.yaml"), 'r') as f:
        moveit_controllers.update({"moveit_simple_controller_manager": yaml.safe_load(f)})

    trajectory_execution = {
        "moveit_manage_controllers": True,
        "trajectory_execution.allowed_execution_duration_scaling": 1.2,
        "trajectory_execution.allowed_goal_duration_margin": 0.5,
        "trajectory_execution.allowed_start_tolerance": 0.01,
    }
    planning_scene_monitor_parameters = {
        "publish_planning_scene": True,
        "publish_geometry_updates": True,
        "publish_state_updates": True,
        "publish_transforms_updates": True,
    }

    return LaunchDescription([
        # RViz
        Node(
            package='rviz2',
            executable='rviz2',
            output='log',
            arguments=['-d', join(launch_root, 'rviz', 'ar3.rviz')],
            parameters=[robot_description, robot_description_semantic]),

        # Attatch robot to map
        Node(
            package='tf2_ros',
            executable='static_transform_publisher',
            output='screen',
            arguments=['0', '0', '0', '0', '0', '0', 'world', 'base_link']),

        # Controller
        Node(
            package='controller_manager',
            executable='ros2_control_node',
            output='screen',
            parameters=[robot_description, join(config_root, 'controller_manager.yaml')]),

        # Start controllers
        ExecuteProcess(
            cmd=['ros2 run controller_manager spawner.py joint_trajectory_controller'],
            shell=True,
            output='screen'),
        ExecuteProcess(
            cmd=['ros2 run controller_manager spawner.py joint_state_controller'],
            shell=True,
            output='screen'),

        # Start the actual move_group node/action server
        Node(
            package="moveit_ros_move_group",
            executable="move_group",
            output="screen",
            parameters=[
                robot_description,
                robot_description_semantic,
                robot_description_kinematics,
                ompl_planning_pipeline_config,
                trajectory_execution,
                moveit_controllers,
                planning_scene_monitor_parameters
            ],
        ),

        # Robot state publisher
        Node(
            package='robot_state_publisher',
            executable='robot_state_publisher',
            output='screen',
            parameters=[robot_description]),

        # Fake robot state publisher
        Node(
            package='ar3_description',
            executable='state_publisher',
            name='state_publisher',
            emulate_tty=True,
            output='screen'),
    ])
