#include "ar3_hardware/ar3_system.hpp"

#include <chrono>
#include <cmath>
#include <limits>
#include <memory>
#include <vector>

#include "hardware_interface/types/hardware_interface_type_values.hpp"

namespace ar3_hardware
{

hardware_interface::return_type Ar3SystemHardware::configure(
  const hardware_interface::HardwareInfo & info)
{
  if (configure_default(info) != hardware_interface::return_type::OK)
  {
    return hardware_interface::return_type::ERROR;
  }

  //hw_start_sec_ = stod(info_.hardware_parameters["example_param_hw_start_duration_sec"]);
  //hw_stop_sec_ = stod(info_.hardware_parameters["example_param_hw_stop_duration_sec"]);
  hw_positions_.resize(info_.joints.size(), std::numeric_limits<double>::quiet_NaN());
  hw_commands_.resize(info_.joints.size(), std::numeric_limits<double>::quiet_NaN());

  for (const hardware_interface::ComponentInfo & joint : info_.joints)
  {
    // DiffBotSystem has exactly two states and one command interface on each joint
    if (joint.command_interfaces.size() != 1)
    {
      RCLCPP_FATAL(
        rclcpp::get_logger("Ar3SystemHardware"),
        "Joint '%s' has %d command interfaces found. 1 expected.", joint.name.c_str(),
        joint.command_interfaces.size());
      return hardware_interface::return_type::ERROR;
    }

    if (joint.command_interfaces[0].name != hardware_interface::HW_IF_POSITION)
    {
      RCLCPP_FATAL(
        rclcpp::get_logger("Ar3SystemHardware"),
        "Joint '%s' have %s command interfaces found. '%s' expected.", joint.name.c_str(),
        joint.command_interfaces[0].name.c_str(), hardware_interface::HW_IF_POSITION);
      return hardware_interface::return_type::ERROR;
    }

    if (joint.state_interfaces.size() != 1)
    {
      RCLCPP_FATAL(
        rclcpp::get_logger("Ar3SystemHardware"),
        "Joint '%s' has %d state interface. 1 expected.", joint.name.c_str(),
        joint.state_interfaces.size());
      return hardware_interface::return_type::ERROR;
    }

    if (joint.state_interfaces[0].name != hardware_interface::HW_IF_POSITION)
    {
      RCLCPP_FATAL(
        rclcpp::get_logger("Ar3SystemHardware"),
        "Joint '%s' have '%s' as state interface. '%s' expected.",
        joint.name.c_str(), joint.state_interfaces[0].name.c_str(),
        hardware_interface::HW_IF_POSITION);
      return hardware_interface::return_type::ERROR;
    }
  }

  status_ = hardware_interface::status::CONFIGURED;
  return hardware_interface::return_type::OK;
}

std::vector<hardware_interface::StateInterface> Ar3SystemHardware::export_state_interfaces()
{
  std::vector<hardware_interface::StateInterface> state_interfaces;
  for (auto i = 0u; i < info_.joints.size(); i++)
  {
    state_interfaces.emplace_back(hardware_interface::StateInterface(
      info_.joints[i].name, hardware_interface::HW_IF_POSITION, &hw_positions_[i]));
  }

  return state_interfaces;
}

std::vector<hardware_interface::CommandInterface> Ar3SystemHardware::export_command_interfaces()
{
  std::vector<hardware_interface::CommandInterface> command_interfaces;
  for (auto i = 0u; i < info_.joints.size(); i++)
  {
    command_interfaces.emplace_back(hardware_interface::CommandInterface(
      info_.joints[i].name, hardware_interface::HW_IF_POSITION, &hw_commands_[i]));
  }

  return command_interfaces;
}

hardware_interface::return_type Ar3SystemHardware::start()
{
  RCLCPP_INFO(rclcpp::get_logger("Ar3SystemHardware"), "Starting ...please wait...");

  // set some default values
  for (auto i = 0u; i < hw_positions_.size(); i++)
  {
    if (std::isnan(hw_positions_[i]))
    {
      hw_positions_[i] = 0;
      hw_commands_[i] = 0;
    }
  }

  executor_ = std::make_shared<rclcpp::executors::MultiThreadedExecutor>();
  nh_ = std::make_shared<rclcpp::Node>("ar3_hardware");
  executor_->add_node(nh_);

  auto spin = [this]() {
    while(rclcpp::ok()) {
      executor_->spin_once();
    }
  };
  thread_executor_spin_ = std::thread(spin);

  j1_pub_ = nh_->create_publisher<std_msgs::msg::Float32>("/net_robot/drives/j1", 10);
  j2_pub_ = nh_->create_publisher<std_msgs::msg::Float32>("/net_robot/drives/j2", 10);
  j3_pub_ = nh_->create_publisher<std_msgs::msg::Float32>("/net_robot/drives/j3", 10);
  j4_pub_ = nh_->create_publisher<std_msgs::msg::Float32>("/net_robot/drives/j4", 10);
  j5_pub_ = nh_->create_publisher<std_msgs::msg::Float32>("/net_robot/drives/j5", 10);
  j6_pub_ = nh_->create_publisher<std_msgs::msg::Float32>("/net_robot/drives/j6", 10);

  status_ = hardware_interface::status::STARTED;

  RCLCPP_INFO(rclcpp::get_logger("Ar3SystemHardware"), "System Successfully started!");

  return hardware_interface::return_type::OK;
}

hardware_interface::return_type Ar3SystemHardware::stop()
{
  RCLCPP_INFO(rclcpp::get_logger("Ar3SystemHardware"), "Stopping ...please wait...");

  status_ = hardware_interface::status::STOPPED;

  RCLCPP_INFO(rclcpp::get_logger("Ar3SystemHardware"), "System successfully stopped!");

  return hardware_interface::return_type::OK;
}

hardware_interface::return_type Ar3SystemHardware::read()
{
  //RCLCPP_INFO(rclcpp::get_logger("Ar3SystemHardware"), "Reading...");

  for (uint i = 0; i < hw_commands_.size(); i++)
  {
    hw_positions_[i] = 6.9;

  //  RCLCPP_INFO(
  //    rclcpp::get_logger("Ar3SystemHardware"),
  //    "Got position state %.5f for '%s'!", hw_positions_[i],
  //    info_.joints[i].name.c_str());
  }

  //RCLCPP_INFO(
  //  rclcpp::get_logger("Ar3SystemHardware"), "Joints successfully read! (%.5f,%.5f,%.5f)",
  //  hw_commands_[0], hw_commands_[1], hw_commands_[2]);

  return hardware_interface::return_type::OK;
}

hardware_interface::return_type ar3_hardware::Ar3SystemHardware::write()
{
  //RCLCPP_INFO(rclcpp::get_logger("Ar3SystemHardware"), "Writing...");

  auto message = std_msgs::msg::Float32();
  message.data = hw_commands_[0];
  j1_pub_->publish(message);
  message.data = hw_commands_[1];
  j2_pub_->publish(message);
  message.data = hw_commands_[2];
  j3_pub_->publish(message);
  message.data = hw_commands_[3];
  j4_pub_->publish(message);
  message.data = hw_commands_[4];
  j5_pub_->publish(message);
  message.data = hw_commands_[5];
  j6_pub_->publish(message);

  //for (auto i = 0u; i < hw_commands_.size(); i++)
  //{
    // Simulate sending commands to the hardware
  //  RCLCPP_INFO(
  //    rclcpp::get_logger("Ar3SystemHardware"), "Got command %.5f for '%s'!", hw_commands_[i],
  //    info_.joints[i].name.c_str());
  //}
  //RCLCPP_INFO(rclcpp::get_logger("Ar3SystemHardware"), "Joints successfully written!");

  return hardware_interface::return_type::OK;
}

}  // namespace ar3_hardware

#include "pluginlib/class_list_macros.hpp"
PLUGINLIB_EXPORT_CLASS(
  ar3_hardware::Ar3SystemHardware, hardware_interface::SystemInterface)
